from nautobot.dcim.choices import CableStatusChoices
from nautobot.extras.plugins import PluginTemplateExtension

from .models import MeasurementLog
from .models import CableTemplate


class CableCommission(PluginTemplateExtension):
    model = 'dcim.cable'

    def buttons(self):
        cable = self.context['object']
        log = None
        cable_template = None

        maybe_log = MeasurementLog.objects.filter(cable=cable)
        if maybe_log.exists():
            log = maybe_log.first().link

        cable_template = CableTemplate.objects.filter(cable=cable)

        return self.render('nautobot_cable_utils/inc/buttons.html', {
            'cable': cable,
            'cable_planned': cable.status.slug == CableStatusChoices.STATUS_PLANNED,
            'cable_inventory': cable_template.exists(),
            'log': log,
        })


class DeviceBulkConnect(PluginTemplateExtension):
    model = 'dcim.device'

    def buttons(self):
        device = self.context['object']
        cable_available = None

        if device.get_cables().filter(status__slug="planned").count() > 0:
            cable_available = device.get_cables().filter(status__slug="planned").first().pk

        return self.render('nautobot_cable_utils/inc/buttons_device.html', {
            'device': device,
            'cable_planned': cable_available != None,
        })


class InterfaceAutoRouteStart(PluginTemplateExtension):
    model = "dcim.interface"

    def buttons(self):
        return self.render(
            'nautobot_cable_utils/inc/buttons_interface.html',
        )


class FrontPortAutoRouteStart(PluginTemplateExtension):
    model = "dcim.frontport"

    def buttons(self):
        return self.render(
            'nautobot_cable_utils/inc/buttons_frontport.html',
        )


template_extensions = [CableCommission, DeviceBulkConnect, InterfaceAutoRouteStart, FrontPortAutoRouteStart]



# Generated by Django 3.2.13 on 2022-08-08 09:42

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('nautobot_cable_utils', '0002_auto_20220808_0747'),
    ]

    operations = [
        migrations.RenameField(
            model_name='cabletemplate',
            old_name='tenant',
            new_name='owner',
        ),
    ]



from nautobot.extras.plugins import PluginConfig

class CableUtilsConfig(PluginConfig):
    name = 'nautobot_cable_utils'
    verbose_name = 'Cable Utilities'
    description = 'A cable utilities plugin'
    version = "0.8.0"
    author = "Gesellschaft für wissenschaftliche Datenverarbeitung mbH Göttingen"
    author_email = "netzadmin@gwdg.de"
    base_url = 'nautobot-cable-utils'
    required_settings = [
    ]
    default_settings = {
    }
    middleware = [
    ]

config = CableUtilsConfig



from nautobot.extras.plugins import PluginMenuButton, PluginMenuItem
from nautobot.utilities.choices import ButtonColorChoices

menu_items = (
    PluginMenuItem(
        link='plugins:nautobot_cable_utils:cabletemplate_list',
        link_text='Cable Templates',
        buttons=(
            PluginMenuButton('plugins:nautobot_cable_utils:cabletemplate_add', 'Add Cable Template', 'mdi mdi-plus-thick', ButtonColorChoices.GREEN),
            PluginMenuButton('plugins:nautobot_cable_utils:cabletemplate_import', 'Import Cable Templates', 'mdi mdi-database-import-outline', ButtonColorChoices.BLUE),
        )
    ),
    PluginMenuItem(
        link='plugins:nautobot_cable_utils:measurement_log_list',
        link_text='Measurement Logs',
        buttons=(
            PluginMenuButton('plugins:nautobot_cable_utils:measurement_log_add', 'Add Measurement Log', 'mdi mdi-plus-thick', ButtonColorChoices.GREEN),
        )
    ),
)



from django.contrib import messages
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404, redirect, render
from django.urls import reverse
from django.views.generic import View, FormView

from nautobot.core.views import generic
from nautobot.dcim.choices import CableStatusChoices
from nautobot.dcim.models import RearPort, Cable, Device, Interface, FrontPort
from nautobot.core.views.generic import (
    BulkDeleteView, BulkImportView, GetReturnURLMixin, ObjectEditView,
    ObjectListView
)
from nautobot.utilities.forms import ConfirmationForm
from nautobot.extras.models.statuses import Status

import time

from . import filters, forms, tables
from .cable_router import CableRouter, CablePath
from .forms import CableRouterVerifyForm, CableRouterLinkToInterfaceForm, \
    CableRouterLinkToFrontPortForm
from .models import CableTemplate, MeasurementLog
from .tables import NeededCableTable


class ReconnectView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.add_cable'
    template_name = 'nautobot_cable_utils/cable_connect.html'

    def dispatch(self, request, *args, pk=None, **kwargs):
        self.obj = Cable.objects.get(pk=pk)

        idx = (self.obj.termination_a_type.model, self.obj.termination_b_type.model)
        idx_sorted = tuple(sorted(idx))
        if idx != idx_sorted:
            termination_a = self.obj.termination_a
            self.obj.termination_a = self.obj.termination_b
            self.obj.termination_b = termination_a

        self.form_class = {
            ('circuittermination', 'circuittermination'): forms.ConnectCircuitTerminationForm,
            ('circuittermination', 'rearport'): forms.ConnectCircuitTerminationToRearPortForm,
            ('consoleport', 'consoleserverport'): forms.ConnectConsolePortToConsoleServerPortForm,
            ('consoleport', 'frontport'): forms.ConnectConsolePortToFrontPortForm,
            ('consoleport', 'rearport'): forms.ConnectConsolePortToRearPortForm,
            ('consoleserverport', 'frontport'): forms.ConnectConsoleServerPortToFrontPortForm,
            ('consoleserverport', 'rearport'): forms.ConnectConsoleServerPortToRearPortForm,
            ('powerfeed', 'powerport'): forms.ConnectPowerfeedToPowerPortForm,
            ('poweroutlet', 'powerport'): forms.ConnectPowerOutletToPowerPortForm,
            ('circuittermination', 'interface'): forms.ConnectCircuitTerminationToInterfaceForm,
            ('frontport', 'interface'): forms.ConnectFrontPortToInterfaceForm,
            ('interface', 'rearport'): forms.ConnectInterfaceToRearPortForm,
            ('interface', 'interface'): forms.ConnectInterfaceForm,
            ('frontport', 'frontport'): forms.ConnectFrontPortForm,
            ('frontport', 'rearport'): forms.ConnectFrontPortToRearPortForm,
            ('circuittermination', 'frontport'): forms.ConnectCircuitTerminationToFrontPortForm,
            ('rearport', 'rearport'): forms.ConnectRearPortForm,
        }[idx_sorted]

        return super().dispatch(request, *args, **kwargs)

    def prefill_form(self, initial_data, termination):
        o = getattr(self.obj, termination)
        if o and hasattr(o, 'device'):
            device = o.device
            initial_data['{}_device'.format(termination)] = device
            if device.site:
                initial_data['{}_site'.format(termination)] = device.site
            if device.rack:
                initial_data['{}_rack'.format(termination)] = device.rack

    def get(self, request, *args, **kwargs):
        # Parse initial data manually to avoid setting field values as lists
        initial_data = {k: request.GET[k] for k in request.GET}

        self.prefill_form(initial_data, 'termination_a')
        self.prefill_form(initial_data, 'termination_b')

        form = self.form_class(instance=self.obj, initial=initial_data)

        return render(request, self.template_name, {
            'obj': self.obj,
            'obj_type': Cable._meta.verbose_name,
            'form': form,
            'return_url': self.get_return_url(request, self.obj),
        })

    def post(self, request, *args, **kwargs):
        assigned_cable_template = None

        with transaction.atomic():
            try:
                if hasattr(self.obj.termination_a, '_path') and self.obj.termination_a._path_id:
                    self.obj.termination_a._path = None
            except CablePath.DoesNotExist:
                pass
            try:
                if hasattr(self.obj.termination_b, '_path') and self.obj.termination_b._path_id:
                    self.obj.termination_b._path = None
            except CablePath.DoesNotExist:
                pass

            try:
                assigned_cable_template = CableTemplate.objects.get(cable=self.obj)
            except:
                pass

            self.obj.delete()
            self.obj.pk = None
            self.obj._state.adding = True

            self.obj.termination_a.cable = None
            self.obj.termination_b.cable = None

            form = self.form_class(request.POST, request.FILES, instance=self.obj)
            if form.is_valid():
                obj = form.save()

                if assigned_cable_template != None:
                    # restore cable template if previously assigned
                    assigned_cable_template.cable = obj
                    assigned_cable_template.save()

                redirect_url = self.get_return_url(request, obj)
                return redirect(redirect_url)

            return render(request, self.template_name, {
                'obj': self.obj,
                'obj_type': Cable._meta.verbose_name,
                'form': form,
                'return_url': self.get_return_url(request, self.obj),
            })


class CommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.edit_cable'
    template_name = 'nautobot_cable_utils/commission_cable.html'

    def get(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = forms.CommissionForm(initial={'cable': cable})

        return render(request, self.template_name, {
            'form': form,
            'cable': cable,
            'return_url': self.get_return_url(request, cable),
        })

    def post(self, request, pk=None):
        with transaction.atomic():
            cable = get_object_or_404(Cable, pk=pk)
            form = forms.CommissionForm(request.POST, initial={'cable': cable})

            if form.is_valid():
                form.save()

                return redirect(self.get_return_url(request, cable))

            return render(request, self.template_name, {
                'form': form,
                'cable': cable,
                'return_url': self.get_return_url(request, cable),
            })


class UndoCommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.edit_cable'
    queryset = Cable.objects.all()
    template_name = 'nautobot_cable_utils/undo_commission.html'

    def get(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = ConfirmationForm(initial=request.GET)

        return render(
            request,
            self.template_name,
            {
                "obj": cable,
                "form": form,
                "obj_type": self.queryset.model._meta.verbose_name,
                "return_url": self.get_return_url(request, cable),
            },
        )

    def post(self, request, pk=None):
        cable = get_object_or_404(Cable, pk=pk)
        form = ConfirmationForm(request.POST)

        if form.is_valid():
            with transaction.atomic():
                cable_new = Cable()
                cable_new.termination_a = cable.termination_a
                cable_new.termination_b = cable.termination_b
                cable_new.status = Status.objects.get(slug="planned")

                cable.delete()
                cable_new.save()

        return redirect(self.get_return_url(request, cable_new))


class DeviceCommissionView(PermissionRequiredMixin, GetReturnURLMixin, View):
    permission_required = 'dcim.edit_cable'
    template_name = 'nautobot_cable_utils/bulk_commission.html'

    def get(self, request, pk=None):
        device = get_object_or_404(Device, pk=pk)

        # set key, if not present
        if 'current_device' not in request.session:
            request.session['current_device'] = None

        if request.session['current_device'] != str(pk):
            cable_list = list()
            for cable in device.get_cables().filter(status__slug="planned"):
                cable_list.append(str(cable.pk))

            request.session['current_device'] = str(pk)
            request.session['cable_list'] = cable_list
            request.session.modified = True

        cable = get_object_or_404(Cable, pk=request.session['cable_list'][0])
        form = forms.CommissionForm(initial={'cable': cable})

        return render(request, self.template_name, {
            'form': form,
            'count': len(request.session['cable_list']),
            'object': cable,
            'verbose_name': Cable._meta.verbose_name,
            'verbose_name_plural': Cable._meta.verbose_name_plural,
            'return_url': self.get_return_url(request, device),
            'active_tab': 'main',
        })

    def post(self, request, pk=None):
        with transaction.atomic():
            device = get_object_or_404(Device, pk=pk)
            cable = get_object_or_404(Cable, pk=request.session['cable_list'][0])
            form = forms.CommissionForm(request.POST, initial={'cable': cable})

            if '_create' in self.request.POST:
                if form.is_valid():
                    request.session['cable_list'].pop(0)
                    request.session.modified = True
                    form.save()
            elif '_skip' in self.request.POST:
                request.session['cable_list'].pop(0)
                request.session.modified = True

            if len(request.session['cable_list']) > 0 and '_cancel' not in self.request.POST:
                if form.is_valid() or '_skip' in self.request.POST:
                    return redirect(request.path)
                else:
                    # re-render page to display ValidationErrors
                    return render(request, self.template_name, {
                        'form': form,
                        'count': len(request.session['cable_list']),
                        'object': get_object_or_404(Cable, pk=request.session['cable_list'][0]),
                        'verbose_name': Cable._meta.verbose_name,
                        'verbose_name_plural': Cable._meta.verbose_name_plural,
                        'return_url': self.get_return_url(request, device),
                        'active_tab': 'main',
                    })

            if 'current_device' in self.request.session.keys():
                del request.session['current_device']
            if 'cable_list' in self.request.session.keys():
                del request.session['cable_list']

        return redirect(self.get_return_url(request, device))


class CableTemplateListView(PermissionRequiredMixin, ObjectListView):
    permission_required = 'dcim.view_cable_template'
    queryset = CableTemplate.objects.all()
    filterset = filters.CableTemplateFilterSet
    filterset_form = forms.CableTemplateFilterForm
    table = tables.CableTemplateTable
    action_buttons = ("add", "import")
    template_name = 'nautobot_cable_utils/cable_template_list.html'


class CableTemplateCreateView(PermissionRequiredMixin, ObjectEditView):
    permission_required = 'dcim.add_cable_template'
    queryset = CableTemplate.objects.all()
    model_form = forms.CableTemplateForm
    default_return_url = 'plugins:nautobot_cable_utils:cabletemplate_list'
    template_name = "nautobot_cable_utils/cabletemplate_edit.html"


class CableTemplateBulkCreateView(generic.BulkCreateView):
    queryset = CableTemplate.objects.all()
    form = forms.CableTemplateBulkCreateForm
    model_form = forms.CableTemplateBulkForm
    pattern_target = "name"
    template_name = "nautobot_cable_utils/cabletemplate_bulk_add.html"


class CableTemplateEditView(CableTemplateCreateView):
    permission_required = 'dcim.edit_cable_template'


class CableTemplateBulkImportView(BulkImportView):
    queryset = CableTemplate.objects.all()
    model_form = forms.CableTemplateCSVForm
    table = tables.CableTemplateTable


class CableTemplateBulkDeleteView(BulkDeleteView):
    queryset = CableTemplate.objects.all()
    filterset = filters.CableTemplateFilterSet
    table = tables.CableTemplateTable


class MeasurementLogListView(PermissionRequiredMixin, ObjectListView):
    permission_required = 'dcim.view_measurement_log'
    queryset = MeasurementLog.objects.all()
    table = tables.MeasurementLogTable
    action_buttons = tuple()
    template_name = 'nautobot_cable_utils/measurement_log_list.html'


class MeasurementLogCreateView(PermissionRequiredMixin, ObjectEditView):
    permission_required = 'dcim.add_measurement_log'
    queryset = MeasurementLog.objects.all()
    model_form = forms.MeasurementLogForm
    default_return_url = 'plugins:nautobot_cable_utils:measurement_log_list'


class MeasurementLogEditView(CableTemplateCreateView):
    permission_required = 'dcim.edit_measurement_log'


class MeasurementLogBulkDeleteView(BulkDeleteView):
    queryset = MeasurementLog.objects.all()
    table = tables.MeasurementLogTable


class CableRouterAddView(PermissionRequiredMixin, FormView):
    permission_required = 'dcim.add_cable'
    template_name = "nautobot_cable_utils/cable_router_add.html"

    termination_types = {
        "interface": ContentType.objects.get_for_model(Interface),
        "front-port": ContentType.objects.get_for_model(FrontPort),
    }

    termination_forms = {
        "interface": CableRouterLinkToInterfaceForm,
        "front-port": CableRouterLinkToFrontPortForm,
    }

    termination_names = {
        'interface': "Interface",
        "front-port": "Front Port",
    }

    termination_a_type = None
    termination_b_type = None
    termination_a_type_name = None
    termination_b_type_name = None
    termination_a = None
    form_class = None

    def dispatch(self, request, *args, **kwargs):
        self.termination_a_type = self.termination_types[kwargs.get("termination_a_type")]
        self.termination_b_type = self.termination_types[kwargs.get("termination_b_type")]
        self.termination_a_type_name = self.termination_names[kwargs.get("termination_a_type")]
        self.termination_b_type_name = self.termination_names[kwargs.get("termination_b_type")]
        self.termination_a = self.termination_a_type.model_class().objects.get(id=kwargs.get("termination_a_id"))
        self.form_class = self.termination_forms[kwargs.get("termination_b_type")]

        return super().dispatch(request, *args, **kwargs)

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        return {
            **kwargs,
            "termination_a": self.termination_a,
            "termination_a_type": self.termination_a_type,
            "termination_b_type": self.termination_b_type,
        }

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data()

        return {
            **ctx,
            "termination_a_type": self.termination_a_type,
            "termination_b_type": self.termination_b_type,
            "termination_a_type_name": self.termination_a_type_name,
            "termination_b_type_name": self.termination_b_type_name,
            "termination_a": self.termination_a,
        }

    def form_valid(self, form):
        query_params1 = map(lambda cid: f"cable={cid}", form.cleaned_data['path'])
        query_params2 = [
            f"media_type={form.cleaned_data['media_type']}",
            f"termination_a_type={self.termination_a_type.id}",
            f"termination_a_id={self.termination_a.id}",
            f"termination_b_type={self.termination_b_type.id}",
            f"termination_b_id={form.cleaned_data['termination_b_id']}"
        ]
        query_params = "&".join([*query_params1, *query_params2])
        redirect_url = reverse("plugins:nautobot_cable_utils:cable_router_verify") + "?" + query_params
        return HttpResponseRedirect(redirect_url)


class CableRouterVerifyView(PermissionRequiredMixin, FormView):
    permission_required = 'dcim.add_cable'
    form_class = CableRouterVerifyForm
    template_name = "nautobot_cable_utils/cable_router_verify.html"

    def get_cable_path(self):
        ct_termination_a = ContentType.objects.get_for_id(self.request.GET.get("termination_a_type"))
        ct_termination_b = ContentType.objects.get_for_id(self.request.GET.get("termination_b_type"))

        termination_a = ct_termination_a.model_class().objects.get(pk=self.request.GET.get("termination_a_id"))
        termination_b = ct_termination_b.model_class().objects.get(pk=self.request.GET.get("termination_b_id"))

        cable_list = self.request.GET.getlist("cable")
        cable_path = CablePath(termination_a, ct_termination_a, termination_b, ct_termination_b, cable_list)

        return cable_path

    def get_context_data(self, **kwargs):
        ctx = super().get_context_data(**kwargs)
        cable_path = self.get_cable_path()

        needed_cable_table = NeededCableTable(cable_path.get_needed_cables())

        return {
            **ctx,
            'termination_a': cable_path.termination_a,
            "termination_b": cable_path.termination_b,
            "path": cable_path.get_renderable_path(),
            "needed_cable_table": needed_cable_table,
        }

    def form_valid(self, form):
        cable_path = self.get_cable_path()
        needed_cables = cable_path.get_needed_cables()
        planned_status = Status.objects.get(slug=CableStatusChoices.STATUS_PLANNED)

        for needed_cable in needed_cables:
            Cable.objects.create(
                termination_a=needed_cable['termination_a'],
                termination_b=needed_cable['termination_b'],
                status=planned_status,
            )

        messages.success(self.request, f"{len(needed_cables)} Cables were created.")

        return HttpResponseRedirect(cable_path.termination_a.get_absolute_url())



import re
from uuid import UUID

from django.contrib.contenttypes.models import ContentType
from django.forms import ValidationError

from igraph import Graph
from nautobot.dcim.choices import CableTypeChoices
from nautobot.dcim.models import Device, FrontPort, RearPort, Rack, Cable


def get_weight_for_frontport_name(fp_name):
    if "/" in fp_name:
        [part_1, part_2] = fp_name.split("/")
        part_1_match = re.search(r'\d+', part_1)
        part_2_match = re.search(r'\d+', part_2)

        part_1_num = int(part_1_match.group(0)) if part_1_match else 0
        part_2_num = int(part_2_match.group(0)) if part_2_match else 0
        return 10000 + (part_1_num * 1000) + part_2_num

    else:
        part_1_match = re.search(r'\d+', fp_name)
        part_1_num = int(part_1_match.group(0)) if part_1_match else 0
        return 10000 + part_1_num


class CablePath:
    def __init__(self, termination_a, termination_a_type, termination_b, termination_b_type, cable_inp):
        self.termination_a = termination_a
        self.termination_b = termination_b
        self.termination_a_type = termination_a_type
        self.termination_b_type = termination_b_type

        rack_list = [self.termination_a.device.rack]
        cable_list = []

        cables_to_pop = [*cable_inp]
        cable_objs = Cable.objects.filter(id__in=cable_inp).prefetch_related("_termination_a_device__rack",
                                                                             "_termination_b_device__rack")
        while len(cables_to_pop) > 0:

            next_cable = None
            is_swapped = False
            for c_id in cables_to_pop:
                c = next(filter(lambda co: co.id == UUID(c_id), cable_objs))
                if c.termination_a.device.rack == rack_list[-1]:
                    next_cable = c
                    is_swapped = False
                elif c.termination_b.device.rack == rack_list[-1]:
                    next_cable = c
                    is_swapped = True

            cable_list.append(next_cable)
            rack_list.append(
                next_cable.termination_a.device.rack if is_swapped else next_cable.termination_b.device.rack)
            cables_to_pop.pop(cables_to_pop.index(str(next_cable.id)))

        self.rack_list = rack_list
        self.cable_list = cable_list

    def get_renderable_path(self):
        needed_cables = self.get_needed_cables()

        # needed_cables and self.cable_list are sorted and only needs to be zipped for a complete cable trace.
        # Afterwards, we only need to add interfaces and devices for a neat trace.

        intermediate_elements = list()

        open_element = {
            "device": self.termination_a.device,
        }

        # We use UUIDs, so we do only compare for the same UUID, we do not care about the type, should fit anyways.

        cable_list = [*self.cable_list]

        if self.termination_a_type.app_label == "dcim" and self.termination_a_type.model == "interface":
            fp_to_interface_cable = needed_cables.pop(0)

            is_swapped = fp_to_interface_cable['termination_b'].device == open_element['device']
            intermediate_elements.append({
                **open_element,
                "attachment_b": fp_to_interface_cable['termination_b'] if is_swapped else fp_to_interface_cable[
                    'termination_a'],
            })

            intermediate_elements.append({
                "needed_cable": fp_to_interface_cable,
            })

            next_attachment_point = fp_to_interface_cable['termination_b'] if not is_swapped else fp_to_interface_cable[
                'termination_a']

            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }
        elif self.termination_a_type.app_label == "dcim" and self.termination_a_type.model == "frontport":
            open_element["attachment_a"] = self.termination_a

        prepended_cable = None
        if self.termination_b_type.app_label == "dcim" and self.termination_b_type.model == "interface":
            prepended_cable = needed_cables.pop()

        for i, next_existing_cable in enumerate(cable_list):

            is_swapped = next_existing_cable.termination_b.device == open_element['device']

            open_element["attachment_b"] = next_existing_cable.termination_b if is_swapped else \
                next_existing_cable.termination_a

            intermediate_elements.append(open_element)

            intermediate_elements.append({
                "cable": next_existing_cable
            })

            next_attachment_point = next_existing_cable.termination_a if is_swapped else next_existing_cable.termination_b
            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }

            if i >= len(needed_cables):
                continue

            needed_cable = needed_cables[i]

            is_swapped = needed_cable['termination_b'].device == open_element['device']
            intermediate_elements.append({
                **open_element,
                "attachment_b": needed_cable['termination_b'] if is_swapped else needed_cable['termination_a'],
            })

            intermediate_elements.append({
                "needed_cable": needed_cable,
            })

            next_attachment_point = needed_cable['termination_b'] if not is_swapped else needed_cable['termination_a']

            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }

        if prepended_cable:

            is_swapped = prepended_cable['termination_b'].device == open_element['device']
            intermediate_elements.append({
                **open_element,
                "attachment_b": prepended_cable['termination_b'] if is_swapped else prepended_cable['termination_a'],
            })

            intermediate_elements.append({
                "needed_cable": prepended_cable,
            })

            next_attachment_point = prepended_cable['termination_b'] if not is_swapped else prepended_cable[
                'termination_a']

            open_element = {
                "attachment_a": next_attachment_point,
                "device": next_attachment_point.device,
            }
        else:
            open_element["attachment_b"] = self.termination_b

        if open_element:
            intermediate_elements.append(open_element)

        return intermediate_elements

    def get_needed_cables(self):

        rearport_uuids = list()
        for c in self.cable_list:
            rearport_uuids.append(c.termination_a_id)
            rearport_uuids.append(c.termination_b_id)

        devices = Device.objects.filter(rearports__in=rearport_uuids)
        needed_cables = list()

        for rack in self.rack_list:

            if rack.id == self.termination_a.device.rack.id and self.termination_a_type.app_label == "dcim" and self.termination_a_type.model == "frontport":
                # We need no first cable to the FP, we ware already at the FP.
                continue
            elif rack.id == self.termination_b.device.rack.id and self.termination_b_type.app_label == "dcim" and self.termination_b_type.model == "frontport":
                # We need no first cable to the FP, we ware already at the FP.
                continue
            elif rack.id == self.termination_a.device.rack.id:
                # Start Rack
                pp_device = next(filter(lambda d: d.rack.id == rack.id, devices))
                fp_to_use = pp_device.frontports.filter(cable=None, rear_port__in=rearport_uuids).first()

                if not fp_to_use:
                    raise ValidationError(f"{pp_device} has no free ports.")

                needed_cables.append({
                    "termination_a_type": ContentType.objects.get_for_model(self.termination_a),
                    "termination_a_id": self.termination_a.id,
                    "termination_a": self.termination_a,
                    "termination_b_type": ContentType.objects.get_for_model(fp_to_use),
                    "termination_b_id": fp_to_use.id,
                    "termination_b": fp_to_use
                })
            elif rack.id == self.termination_b.device.rack.id:
                # Destination Rack
                pp_device = next(filter(lambda d: d.rack.id == rack.id, devices))
                fp_to_use = pp_device.frontports.filter(cable=None, rear_port__in=rearport_uuids).first()

                if not fp_to_use:
                    raise ValidationError(f"{pp_device} has no free ports.")

                needed_cables.append({
                    "termination_a_type": ContentType.objects.get_for_model(fp_to_use),
                    "termination_a_id": fp_to_use.id,
                    "termination_a": fp_to_use,
                    "termination_b_type": ContentType.objects.get_for_model(self.termination_b),
                    "termination_b_id": self.termination_b.id,
                    "termination_b": self.termination_b
                })
            else:
                # Hop Rack
                rack_devices = list(filter(lambda d: d.rack.id == rack.id, devices))
                fps_to_use = FrontPort.objects.filter(cable=None, device__in=rack_devices,
                                                      rear_port__in=rearport_uuids)

                if len(fps_to_use) != 2:
                    raise ValidationError(f"{' or '.join(rack_devices)} have no free ports.")

                fp1_to_use = fps_to_use[0]
                fp2_to_use = fps_to_use[1]

                needed_cables.append({
                    "termination_a_type": ContentType.objects.get_for_model(fp1_to_use),
                    "termination_a_id": fp1_to_use.id,
                    "termination_a": fp1_to_use,
                    "termination_b_type": ContentType.objects.get_for_model(fp2_to_use),
                    "termination_b_id": fp2_to_use.id,
                    "termination_b": fp2_to_use,
                })
        return needed_cables


class CableRouter:

    def __init__(self, termination_a, termination_a_type, termination_b, termination_b_type, media_type):
        self.termination_a = termination_a
        self.termination_b = termination_b

        self.allowed_cable_types = None
        if media_type == "fiber_sm":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_SMF,
                CableTypeChoices.TYPE_SMF_OS1,
                CableTypeChoices.TYPE_SMF_OS2,
            ]
        elif media_type == "fiber_mm":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_MMF,
                CableTypeChoices.TYPE_MMF_OM1,
                CableTypeChoices.TYPE_MMF_OM2,
                CableTypeChoices.TYPE_MMF_OM3,
                CableTypeChoices.TYPE_MMF_OM4,
            ]
        elif media_type == "copper":
            self.allowed_cable_types = [
                CableTypeChoices.TYPE_CAT3,
                CableTypeChoices.TYPE_CAT5,
                CableTypeChoices.TYPE_CAT5E,
                CableTypeChoices.TYPE_CAT6,
                CableTypeChoices.TYPE_CAT6A,
                CableTypeChoices.TYPE_CAT7,
                CableTypeChoices.TYPE_CAT7A,
                CableTypeChoices.TYPE_CAT8,
            ]

        excluded_cable_ids = list()
        if termination_a_type.app_label == "dcim" and termination_a_type.model == "frontport":
            rps_to_exclude = RearPort.objects.filter(device=self.termination_a.device)
            for rp in rps_to_exclude:
                if termination_a.rear_port != rp and rp.cable:
                    excluded_cable_ids.append(rp.cable.id)

        if termination_b_type.app_label == "dcim" and termination_b_type.model == "frontport":
            rps_to_exclude = RearPort.objects.filter(device=self.termination_b.device)
            for rp in rps_to_exclude:
                if termination_b.rear_port != rp and rp.cable:
                    excluded_cable_ids.append(rp.cable.id)

        possible_edges = Cable.objects.raw(
            '''
                SELECT 
                    c.*, d_a.rack_id as termination_a_rack_id, d_b.rack_id as termination_b_rack_id, 
                    fp_a.name as termination_a_name, fp_b.name as termination_b_name 
                FROM dcim_cable c
                LEFT JOIN django_content_type ct_a ON ct_a.id = termination_a_type_id
                LEFT JOIN django_content_type ct_b ON ct_b.id = termination_b_type_id
                LEFT JOIN dcim_rearport rp_a ON rp_a.id = termination_a_id
                LEFT JOIN dcim_rearport rp_b ON rp_b.id = termination_b_id
                INNER JOIN dcim_frontport fp_a ON fp_a.rear_port_id = rp_a.id AND fp_a.cable_id IS NULL
                INNER JOIN dcim_frontport fp_b ON fp_b.rear_port_id = rp_b.id AND fp_b.cable_id IS NULL
                LEFT JOIN dcim_device d_a ON rp_a.device_id = d_a.id
                LEFT JOIN dcim_device d_b ON rp_b.device_id = d_b.id
                WHERE (
                    c.type = ANY(%s) AND
                    NOT c.id = ANY(%s) AND 
                    ct_a.app_label = 'dcim' AND ct_a.model = 'rearport' AND
                    ct_b.app_label = 'dcim' AND ct_b.model = 'rearport' AND
                    fp_a.cable_id IS NULL and fp_b.cable_id IS NULL
                )
            ''',
            [self.allowed_cable_types, excluded_cable_ids]
        )

        self.graph = Graph()
        self.edges = dict()

        verticies = set(str(x) for e in possible_edges for x in [e.termination_a_rack_id, e.termination_b_rack_id])
        for v in verticies:
            self.graph.add_vertex(v)

        for edge in possible_edges:
            fp_a_weight = get_weight_for_frontport_name(edge.termination_a_name)
            fp_b_weight = get_weight_for_frontport_name(edge.termination_b_name)

            if not edge.termination_a_rack_id or not edge.termination_b_rack_id:
                continue

            e_to = self.graph.add_edge(
                str(edge.termination_a_rack_id),
                str(edge.termination_b_rack_id),
                weight=fp_a_weight,
            )
            self.edges[e_to.index] = edge.id
            e_from = self.graph.add_edge(
                str(edge.termination_b_rack_id),
                str(edge.termination_a_rack_id),
                weight=fp_b_weight,
                id=edge.id
            )
            self.edges[e_from.index] = edge.id

    def get_devices_for_hop_list(self, hop_list):
        devices = Device.objects.filter(id__in=hop_list)
        return list(map(lambda h: next(filter(lambda d: d.id == h, devices)), hop_list))

    def get_path(self):

        if self.termination_a.device.rack.id == self.termination_b.device.rack.id:
            raise ValidationError(
                f"{self.termination_a.device} and {self.termination_b.device} are located within the same rack")

        shortest_path = self.graph.get_shortest_paths(
            str(self.termination_a.device.rack.id),
            str(self.termination_b.device.rack.id),
            weights="weight",
            output="epath",
        )

        if len(shortest_path) < 1:
            raise ValidationError(
                f"{self.termination_a.device} and {self.termination_b.device} are not connected"
            )

        return [self.edges[x] for x in shortest_path[0]]



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

with open('README.md', encoding="utf-8") as f:
    long_description = f.read()

setup(
    name='nautobot-cable-utils',
    author='Gesellschaft für wissenschaftliche Datenverarbeitung mbH Göttingen',
    version='0.8.0',
    license='Apache-2.0',
    url='https://gitlab-ce.gwdg.de/gwdg-netz/nautobot-plugins/nautobot-cable-utils/',
    description='A Nautobot plugin for working with cables',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages('.'),
    include_package_data=True,
    install_requires=[
        "igraph"
    ],
    zip_safe=False,
)




from __future__ import unicode_literals

from django.urls import path

from . import views

app_name = 'nautobot_cable_utils'
urlpatterns = [
    path(r'reconnect/<uuid:pk>/edit/', views.ReconnectView.as_view(), name='reconnect'),
    path(r'commission/<uuid:pk>/edit/', views.CommissionView.as_view(), name='commission'),
    path(r'undo_commission/<uuid:pk>/edit/', views.UndoCommissionView.as_view(), name='undo_commission'),
    path(r'device_commission/<uuid:pk>/edit/', views.DeviceCommissionView.as_view(), name='device_commission'),
    path(r'cable_templates/', views.CableTemplateListView.as_view(), name='cabletemplate_list'),
    path(r'cable_templates/add/', views.CableTemplateCreateView.as_view(), name='cabletemplate_add'),
    path(r'cable_templates/bulk-add/', views.CableTemplateBulkCreateView.as_view(), name='cabletemplate_bulk_add'),
    path(r'cable_templates/import/', views.CableTemplateBulkImportView.as_view(), name='cabletemplate_import'),
    path(r'cable_templates/delete/', views.CableTemplateBulkDeleteView.as_view(), name='cabletemplate_bulk_delete'),
    path(r'cable_templates/edit/<uuid:pk>', views.CableTemplateEditView.as_view(), name='cabletemplate_edit'),
    path(r'measurement_logs/', views.MeasurementLogListView.as_view(), name='measurement_log_list'),
    path(r'measurement_logs/add/', views.MeasurementLogCreateView.as_view(), name='measurement_log_add'),
    path(r'measurement_logs/delete/', views.MeasurementLogBulkDeleteView.as_view(), name='measurement_log_bulk_delete'),
    path(r'measurement_logs/edit/<uuid:pk>', views.MeasurementLogEditView.as_view(), name='measurement_log_edit'),
    path(r'cable_router/link/<str:termination_a_type>/<uuid:termination_a_id>/to/<str:termination_b_type>', views.CableRouterAddView.as_view(), name='cable_router_link'),
    path(r'cable_router/verify/', views.CableRouterVerifyView.as_view(), name='cable_router_verify')
]
